# Written by Phil Frost <indigo@bitglue.com>. Public Domain. If you like this
# software please consider giving money to Unununium <http://unununium.org/>.

'''Input module for tla <http://gnuarch.org/>'''

import re, os, sys, time, datetime, itertools
from sets import Set
from optparse import OptionParser

import rcsgraph


unknown_archives = Set()


class TlaError( Exception ):
    pass



class NodeFactory( object ):
    '''Factory to create or return existing node instances for a given name.'''

    def __init__( self ):
        self.known_nodes = {}

    def get( self, nodename, *args, **kwargs ):
        '''Get a node by its fully qualified name.

        If the node has been previously referenced, the existing instance is
        returned. Otherwise, a new instance is returned, created with the
        arguments given.
        '''
        try:
            return self.known_nodes[nodename]
        except KeyError:
            sys.stderr.write( 'adding %s\n' % (nodename,) )
            node = Node( nodename, *args, **kwargs )
            self.known_nodes[nodename] = node
            return node



class Log( dict ):
    '''A parsed log from tla cat-archive-log.

    This is a mapping from headers to values. The attribute `body` contains the
    body of the log.
    '''

    def __init__( self, nodename ):
        super( Log, self ).__init__()
        pipe = os.popen(
            "tla cat-archive-log '%s'" % nodename.replace("'","'\\''") ).__iter__()
        lines = pipe.__iter__()
        header = None
        header_re = re.compile( '([\w-]+):(.*)' )

        for line in lines:
            if not line.strip():
                break
            m = re.match( header_re, line )
            if m:
                header = m.group(1)
                self[header] = m.group(2)
            else:
                try:
                    self[header] += '\n'+line
                except KeyError:
                    break
        self.body = ''.join(lines)
        status = pipe.close()
        if status:
            raise TlaError( 'cat-archive-log failed' )



class Node( rcsgraph.Node ):
    '''Create a new node from a fully qualified revision name.

    Note: generally nodes should not be created directly, but a `NodeFactory`
    used instead.

    Parameters:
    -----------
    - `nodename`: a fully qualified node name. ex: "a@b--c/d--e--0.1--patch-5"
    - `max_unmerges`: if there are more than this many unmerges to a
      node, all unmerges are ignored. This is useful for archives that
      do "log pruning". Default: 25. None to disable.
    '''

    parent_name = continuation_of = None
    new_patches = removed_patches = []

    def __init__( self, nodename, trunk=False, max_unmerges=25 ):
        if trunk:
            self.trunk = True
        self.parent_name = get_parent(nodename)
        try:
            self.label = patch(nodename)
        except ValueError:
            self.label = nodename
        try:
            self.group = fq_version(nodename)
        except ValueError:
            pass

        try:
            log = Log( nodename )
        except TlaError:
            pass
        else:
            self.summary = log.get('Summary', '').strip()
            try:
                date = log['Standard-date']
            except KeyError:
                pass
            else:
                date = time.strptime(date.strip(), '%Y-%m-%d %H:%M:%S GMT')
                self.start_date = datetime.datetime(*date[:6])

            self.new_patches = [p for p in log.get('New-patches','').split() if p != nodename]
            self.removed_patches = log.get('Removed-patches','').split()
            if max_unmerges is not None and len(self.removed_patches) > max_unmerges:
                del self.removed_patches
            self.continuation_of = log.get('Continuation-of')

        super(Node, self).__init__(nodename)



def getstatusoutput( *args ):
    p = os.popen( ' '.join(["'%s'" % tok.replace("'","'\\''") for tok in args]) )
    o = p.read()
    s = p.close()
    return s, o


def patchlevel( nodename ):
    '''Return the patch level (ex: 'patch-3') of a node as an integer.'''
    return int(nodename[nodename.rindex('-')+1:])

def patchtype( nodename ):
    '''Return the patch type (ex: 'patch' or 'base') of a node.'''
    return nodename[nodename.rindex('--')+2:nodename.rindex('-')]

def patch( nodename ):
    '''Return the patch (ex: 'patch-5') of a node.'''
    return nodename[nodename.rindex('--')+2:]

def fq_version( nodename ):
    '''Return the fully qualified version (ex: 'a@b--c/d--e--0.1') of a node.'''
    return nodename[:nodename.rindex('--')]

def archive( nodename ):
    '''Return the archive (the entire part before '/') of a node.'''
    return nodename.split('/',2)[0]


def get_parent( nodename ):
    '''Return the parent of a node.

    First try using tla ancestry-graph -i, which will correctly handle tags,
    version-0 and versionfix versions, and any other weirdnes. Failing that,
    guess by decrementing the patch number and changing "patch" to "base" when
    it hits 0.

    If the parent can not be determined, None is returned. This happens when
    trying to guess the parent of base-0, or if a node name is malformed or not
    a regular revision. Note that the parent of the first revision of a project
    is '(null)'; the parent of '(null)' is None.
    '''

    def guess_parent():
        i = nodename.rindex('--')
        version = nodename[:i]
        patchname, patchlevel = nodename[i+2:].split('-')
        patchlevel = int(patchlevel)
        if patchlevel == 0:
            return None
        patchlevel -= 1
        if patchlevel == 0:
            patchname = 'base'
        ancestor = '%s--%s-%s' % (version, patchname, patchlevel)
        return ancestor

    try:
        archive, _ = nodename.split('/',2)
    except ValueError:
        return None
    if archive in unknown_archives:
        return guess_parent()
    status, ancestor = getstatusoutput( 'tla', 'ancestry-graph', '-i', nodename )
    if status:
        unknown_archives.add( archive )
        return guess_parent()
    else:
        return ancestor.strip()



parser = OptionParser('usage: %prog [...] -i tla [options] [nodes] -o [...]')
parser.add_option('-p', '--parent-depth', type='int', dest='parentlevels',
    help='limit parent arcs to N levels (default 250, -1 for unlimited)', default=250, metavar='N')
parser.add_option('-m', '--merge-depth', type='int', dest='mergelevels',
    help='limit merge arcs to N levels (default 1, -1 for unlimited)', default=1, metavar='N')
parser.add_option('-v', '--one-version', action='store_true', dest='one_version',
    help='limit ancestry to one version', default=False)


def print_help():
    '''Print usage help.'''
    parser.print_help()


def parse_options(args):
    '''Parse command line options to the output module.

    What this returns will later be passed to get_nodes.
    '''
    return parser.parse_args(args)


def get_nodes(options):
    '''Return a list of nodes to graph.

    `options` is the result of calling parse_options earlier.
    '''
    options, source_nodes = options
    node_factory = NodeFactory()
    get_node = node_factory.get


    # create trunk nodes: those requested directly and their parents

    trunk_nodes = Set(itertools.imap(lambda n: get_node(n, trunk=True), source_nodes))
    for node in trunk_nodes.copy():
        for parent_levels in itertools.count():
            if node.parent_name is None:
                break
            try:
                if options.one_version and fq_version(node.parent_name) != node.group:
                    break
            except ValueError:
                pass
            if options.parentlevels >= 0 and parent_levels >= options.parentlevels:
                break

            parent = get_node(node.parent_name, trunk=True)
            node.parent = parent
            if node.parent_name in trunk_nodes:
                break
            trunk_nodes.add(parent)
            node = parent


    # add merges

    for merge_levels in itertools.count():
        if not trunk_nodes:
            break
        if options.mergelevels >= 0 and merge_levels >= options.mergelevels:
            break

        added_nodes = Set()
        for node in trunk_nodes:
            if not node.continuation_of:
                new_patches = Set(itertools.imap(get_node, node.new_patches))
                for new_patch in new_patches.copy():
                    # strip new patches that were new in other new patches. In
                    # other words, indirect merges.
                    new_patches.difference_update(new_patch.new_patches)
                added_nodes.union_update(new_patches)
                node.merged_nodes = new_patches
        resolve_to_top(added_nodes, get_node)
        trunk_nodes = added_nodes


    return [get_node(n) for n in source_nodes]


def resolve_to_top(nodes, get_node):
    '''Add parents for iterable `nodes` until reaching base-0.
    
    Returns a new Set, a superset of nodes including all the nodes that were
    added.
    '''

    new = Set(nodes)
    for node in nodes:
        while True:
            if node.parent_name is None:
                break
            try:
                if fq_version(node.parent_name) != node.group:
                    break
            except ValueError:
                break
            parent = get_node(node.parent_name)
            node.parent = parent
            if node.parent_name in new:
                break
            new.add(parent)
            node = parent
    return new
