# Written by Phil Frost <indigo@bitglue.com>. Public Domain. If you like this
# software please consider giving money to Unununium <http://unununium.org/>.

'''Output module for graphviz <http://www.research.att.com/sw/tools/graphviz/>.'''

import os, sys, itertools, textwrap
from sets import Set

import rcsgraph
from math import pi
from random import random
from optparse import OptionParser

class ColorGenerator( object ):
    '''Generates contrasting colors unique to a key.'''

    def __init__( self, invert ):
        self.hue = 0            # last used hue
        self.colors = {}        # map versions to their colors
        self.invert = invert


    def get_color( self, key, light=False ):
        def next_color():
            self.hue += pi/11 # pi divided by a prime; works rather well
            self.hue %= 1.0
            if .08 < self.hue < .25:
                # if it's around yellow, make it darker
                return self.hue, .7+random()*.3, .45+random()*.25
            else:
                return self.hue, .7+random()*.3, .75+random()*.25
        if key is None:
            return '0,0,.5'
        if key not in self.colors:
            self.colors[key] = next_color()
        colors = self.colors[key]
        if light:
            colors = colors[0], 0.08, 1
        # inverted colors became the default, hence the negation
        if not self.invert:
            h, s, v = colors
            h += 0.5
            if h > 1.0:
                h -= 1.0
            colors = h, v, s
        return "%s,%s,%s" % colors


    __getitem__ = get_color



class DotGraph( rcsgraph.Graph ):

    def __init__(self, nodes, options):
        self.options = options
        super(DotGraph, self).__init__(nodes)
        self.get_color = ColorGenerator( options.invert ).get_color


    def _print( self, string ):
        print >> self.options.outfile, string


    def print_subgraphs( self, all_nodes ):
        # group the nodes by archive, version
        groups = {}
        for node in all_nodes:
            groups.setdefault(node.group, []).append(node)

        for group, nodes in groups.iteritems():
            if not self.no_group and group:
                self._print( '    subgraph "cluster_%s" {' % group )
                if self.options.group:
                    self._print( '      node [group="%s"]' % group )
                self._print( '      label="%s"' % group )
                self._print( '      style=filled' )
                self._print( '      color="%s"' % self.get_color( group, light=True ) )
            self._print( '      node [color="%s"]' % self.get_color(group) )
            for node in nodes:
                if self.options.group_trunk and node.trunk:
                    trunk_group = ',group="trunk"'
                else:
                    trunk_group = ''
                if self.options.highlight_trunk and node.trunk:
                    highlight = ',style="filled",fillcolor="%s"' % (self.get_color(group, light=True),)
                else:
                    highlight = ''
                self._print( '      "%s" [label="%s"%s%s]' % (node, self.make_label(node), highlight, trunk_group) )
            if not self.no_group and group:
                self._print( '    }' )


    def make_label(self, node):
        label = node.label
        if self.options.dates:
            if node.start_date:
                label += '\n'+str(node.start_date)
            if node.end_date:
                label += '\n'+str(node.end_date)
        if self.options.summaries is not None and node.summary:
            summary = textwrap.wrap(node.summary, self.options.summary_width)
            if len(summary) > self.options.summaries:
                del summary[self.options.summaries:]
                summary[-1] += '...'
            label += '\n'+'\n'.join(summary)
        return label.encode('string-escape').replace('"','\\"')


    def print_edges( self, all_nodes ):
        for target in all_nodes:
            for source in target.merged_nodes:
                self.print_edge( source, target )
            for source in target.unmerged_nodes:
                self.print_edge( source, target, {'style':'dashed'} )
            if target.parent is not None:
                if target.parent.group == target.group:
                    self.print_edge( target.parent, target, {'style':'bold', 'weight':self.options.parent_weight} )
                elif self.options.show_continuations:
                    self.print_edge( target.parent, target, {'style':'bold'} )


    def print_edge( self, source, target, attrs=None ):
        attributes = {}
        if attrs:
            attributes.update(attrs)
        attributes['color'] = self.get_color(source.group)
        self._print( '  "%s" -> "%s" [%s]' % (
            source,
            target,
            ','.join(['%s="%s"' % i for i in attributes.iteritems()])
            ) )


    def print_key( self, groups ):
        for group in groups:
            self._print( '"%s" [label="%s", color="%s"]' % (group, group.replace('/','/\\n'), self.get_color(group)) )
        self._print( '"'+'" -> "'.join(groups)+'" [style=invis]' )


    def generate_output( self ):
        all_nodes = self.get_all_nodes()
        self._print( 'digraph {' )
        if not self.options.invert:
            self._print( '  bgcolor=black' )
            self._print( '  fontcolor=white' )
            self._print( '  node [fontcolor=white]' )
            self._print( '  edge [fontcolor=white]' )
        if self.options.concentrate:
            self._print( '  concentrate=true' )
        if not self.options.cluster:
            self._print( '  clusterrank=global' )
        else:
            self._print( '  rankdir=LR' )
        self.print_subgraphs(all_nodes)
        self.print_edges(all_nodes)
        if self.options.key:
            groups = Set(itertools.imap(lambda node: node.group, all_nodes))
            groups.discard(None)
            groups = list(groups)
            groups.sort()
            self.print_key(groups)
        self._print( '}' )



parser = OptionParser('usage: %prog [...] -i [...] -o dot [options]')
parser.add_option('-d', '--dates', action='store_true', dest='dates',
    help='output commit dates on each node', default=False)
parser.add_option('-k', '--key', action='store_true', dest='key',
    help='output a group key', default=False)
parser.add_option('-i', '--invert', action='store_true', dest='invert',
    help='invert colors; white background', default=False)
parser.add_option('--no-concentrate', action='store_false', dest='concentrate',
    help='do not concentrate arcs', default=True)
parser.add_option('--cluster', action='store_true', dest='cluster',
    help='do not cluster nodes of like versions', default=False)
parser.add_option('-f', '--outfile', dest='outfile',
    help='output to FILE (default stdout)', metavar='FILE')
parser.add_option('-w', '--parent-weight', type='int', dest='parent_weight',
    help='weight of direct parent edges. Higher values make branches straighter. (default 1)', metavar='WEIGHT', default=1)
parser.add_option('-g', '--group', action='store_true', dest='group',
    help='use dot\'s "group" attribute to keep branches straight. Similar effects can be made by setting a high weight (-w).', default=False)
parser.add_option('--no-continuations', action='store_false', dest='show_continuations',
    help='omit continuation edges (shows from what revision a branch was tagged)', default=True)
parser.add_option('--no-highlight-trunk', action='store_false', dest='highlight_trunk',
    help='do not highlight trunk nodes', default=True)
parser.add_option('--no-group-trunk', action='store_false', dest='group_trunk',
    help='do not use the group attribute to keep trunk straight', default=True)
parser.add_option('-s', '--summary', type='int', metavar='N', dest='summaries',
    help='show summaries on non-folded nodes, limited to N lines')
parser.add_option('--summary-width', type='int', metavar='N', dest='summary_width',
    help='wrap summary lines at N characters (default 45)', default=45)


def print_help():
    '''Print usage help.'''
    parser.print_help()
    print 'usage: %s [...] -i [...] -o dot' % os.path.basename(sys.argv[0])


def parse_options(args):
    '''Parse command line options to the output module.

    What this returns will later be passed to graph_nodes.
    '''
    options, args = parser.parse_args(args)
    if args:
        print >> sys.stderr, 'unknown option %r' % args[0]
        print >> sys.stderr, 'use "--output-help dot" for help'
        sys.exit(1)
    if options.outfile:
        try:
            options.outfile = file(options.outfile, 'w')
        except IOError, x:
            print >> sys.stderr, 'can not open output file: %s' % x
            sys.exit(1)
    else:
        options.outfile = sys.stdout
    return options


def graph_nodes(nodes, options):
    '''Get the graph for the list of nodes `nodes`

    `options` is the result of calling parse_options earlier.
    '''
    return DotGraph(nodes, options)
